use strict;
use warnings;

package CORBA::XS;

our $VERSION = '0.60';

use CORBA::XS::CdrCVisitor;
use CORBA::XS::CVisitor;
use CORBA::XS::SkeletonCVisitor;
use CORBA::XS::StubCVisitor;
use CORBA::XS::StubPerlVisitor;

1;

